<?php

define("_FC_IMG_PLATFORM_LINUX", "wkhtmltoimage");
define("_FC_IMG_PLATFORM_LINUX_64", "wkhtmltoimage");
define("_FC_IMG_PLATFORM_WINDOWS", "wkhtmltoimage");

class FCImageException extends Exception
{

}

/**
 * @param string $outputFilePath The <b>Full</b> path to the output image. (This <b>WILL BE OVERWRITTEN</b> if exists)
 *                               Please note: PHP should have write permissions jin this
 * @param string $inputString The JSON/XML String of the entire chart object. This is what you will pass to FusionCharts.render ()
 * @param string $swfName The FusionCharts SWF File Path
 * @param array $options A set of options:
 * options ('imageType' => 'png' or 'jpg',
 *          'width' => width in pixels
 *          'height' => height in pixels
 *          'nodelay' => Don't wait for animation to complete (this will speed up script but might cause it to not work)
 *          'quality' => quality from 1 to 100
 * )
 * @throws FCImageException
 * @return bool True if success, false if failure. TODO: Log stderr so users can properly submit bug reports.
 */
function echarts_to_image($outputFilePath, $swfName, $inputString, $height, $width, $options = array())
{

    $jsonFlag = false;
    if($inputString[0] === '<') // Check for the first character
    {
        // do nothing. jsonFlag is set to false anyways
    }
    else if($inputString[0] === "{")
    {
        $jsonFlag = true;
    }
    else
    {
        throw new FCImageException("The input string doesn't seem to be either JSON or XML");
    }

    $fileType = "png";
    if(isset($options['imageType']))
        $fileType = $options['imageType'];

    /*
     * Note: sys_get_temp_dir returns /tmp on Linux (don't know about osx)
     * but on the other hand, it returns C:\user\foo\Local\Temp\ on Windows
     *
     * so we need to add a directory separator on Linux but not on windows
     */
    $separator = "";
    if(DIRECTORY_SEPARATOR === "/") // check if unix system. TODO: There will be better ways to do this
        $separator = DIRECTORY_SEPARATOR;

    $imageFileName = sys_get_temp_dir().$separator."FCImage".rand(0, 1000).'.'.$fileType;

    $cwd = __DIR__; // change working directory to the current script's directory
    $env = array(); // set any environment variables here

    // configure the arguments
    $args = "--format $fileType";

    $args = $args." --crop-w ".($width - 1);
    $args = $args." --crop-h ".($height - 1);

    if(isset($options['nodelay']))
    {
        if($options['nodelay'] === true)
            $args = $args." --javascript-delay 0";
    }
    else
    {
        $args = $args." --javascript-delay 1000";
    }

    if(isset($options['quality']))
    {
        $args = $args." --quality {$options['quality']}";
    }

    $debugFlag = false;
    if(isset($options['debug']))
    {
        $debugFlag = true;
	$debugFile = fopen("debug.html", "w");

	echo "\n\nCall to:\n fusioncharts_to_image ($outputFilePath, $swfName, [removing input string], $height, $width)";
    }

    // now, determine the platform this is running on
    $os = php_uname("s");
    $arch = php_uname("m");
    if($os==="Windows NT")
        $platform = _FC_IMG_PLATFORM_WINDOWS;
    else if($os === "Linux")
    {
        if($arch === "i386")
            $platform = _FC_IMG_PLATFORM_LINUX;
	else if($arch === "i686")
            $platform = _FC_IMG_PLATFORM_LINUX;
        else if ($arch === "x86_64")
            $platform = _FC_IMG_PLATFORM_LINUX_64;
        else
            throw new FCImageException ("This Linux architecture is not supported");
    }
    else
        throw new FCImageException("Your server OS is currently not supported");

    $fcRoot = dirname(__FILE__);

    $wkCommand = $platform;
    $command = "$wkCommand $args - $imageFileName";

    if($debugFlag)
    {
	    echo "\n Command: $command";
    }


    $wkstdin = popen($command, "w");

    if(!is_resource($wkstdin))
    {
        throw new FCImageException("An error took place while trying to open wkhtmltopdf");
    }

    // ok. write template.txt into the process stdin
    fwrite($wkstdin, file_get_contents("$fcRoot/template.txt"));

    if($debugFlag)
    {
        fwrite($debugFile, file_get_contents("$fcRoot/template.txt"));
    }

    $functionToCall = "setXMLData";
    if($jsonFlag === true)
        $functionToCall = "setJSONData";

    // replace all EOL with ""
    // $escapedData = str_replace("\n", "", $inputString);
    // $escapedData = addslashes($escapedData);

// Decode JSON string to PHP array
$array = json_decode($inputString, true);

// Check if decoding was successful
if ($array === null && json_last_error() !== JSON_ERROR_NONE) {
    // Handle error if JSON decoding fails
    echo "JSON decoding failed with error: " . json_last_error_msg();
} else {
    // Print the resulting PHP array
    // print_r($array);

    $arr2 = transformToEChartsOption($array);

    $option = json_encode($arr2);

    echo $option;

}

    $templateCode = <<<BOTTOM

</head>
<body>
<div id="chartContainer" style="height: {$height}px; width: {$width}px;"><small>Loading chart...</small></div>        

<script>
$(function(){

    var dom = document.getElementById('chartContainer');
    var myChart = echarts.init(dom, null, {
    renderer: 'canvas',
    useDirtyRect: false
    });
    var app = {};

    var option = $option;


    console.log("option", option);

    if (option && typeof option === 'object') {
    myChart.setOption(option);
    }

});

</script>
</body>
</html>
BOTTOM;
    fwrite($wkstdin, $templateCode);
    if($debugFlag)
    {
        fwrite($debugFile, $templateCode);
    }

    $returnCode = pclose($wkstdin);

    if($returnCode !== 0)
    {
        if(file_exists($imageFileName))
            unlink($imageFileName);
        throw new FCImageException("There was an error with wkhtmltopdf");
    }

    if (!is_dir($outputFilePath)) {
        if (@mkdir(dirname($outputFilePath), 0777, true)) {
            echo "Created charts folder for this tenant!";
        } else {
            if (is_dir($outputFilePath)) {
                echo "Folder already exists.";
            }
        }
    }

    if (!is_dir($outputFilePath)) {
        if (@mkdir(dirname($outputFilePath), 0777, true)) {
            echo "Created charts folder for this tenant!";
        } else {
            if (is_dir($outputFilePath)) {
                echo "Folder already exists.";
            }
        }
    }

    // success!
    rename($imageFileName, $outputFilePath);

    return true;
}

function transformToEChartsOption($rawData) {
    $data = $rawData['data'] ?? [];
    $chart = $rawData['chart'] ?? [];


    $defaultColors = [
        '#ff600c', '#fcd40e', '#27727B', '#9BCA63', '#C1232B',
        '#61a0a8', '#d48265', '#91c7ae', '#749f83', '#ca8622',
        '#bda29a', '#6e7074', '#546570', '#c4ccd3', '#B5C334',
        '#c23531', '#2f4554', '#61a0a8', '#d48265', '#91c7ae',
        '#749f83', '#ca8622', '#bda29a', '#6e7074', '#546570',
        '#c4ccd3', '#f05b72', '#ef5b9c', '#f47920', '#905a3d',
        '#fab27b', '#2a5caa', '#444693', '#726930', '#b2d235',
        '#6d8346', '#ac6767', '#1d953f', '#6950a1', '#918597',
        '#f6f5ec', '#58b4d0', '#ea66a6', '#ffc773', '#d71345',
        '#ff8c94', '#f2d3ac', '#59678c', '#a4adc0', '#563624',
        '#9d896c', '#c46b35', '#8a5b2f', '#96582a', '#5698c3',
        '#2468a2', '#505e71', '#8d8986', '#d78b2d', '#5e3d50',
        '#4e6e90', '#7d8a8d', '#e0d0b5', '#c0c4c3', '#7d7878'
    ];

    // Precompute color stops
    $colorList = array_map(function($item, $index) use ($defaultColors) {
        $color = strlen($item['color']) > 0 ? '#' . $item['color'] : $defaultColors[$index];
        return [
            'type' => 'linear',
            'x' => 0,
            'y' => 1.5,
            'x2' => 0,
            'y2' => 0,
            'colorStops' => [
                ['offset' => 0, 'color' => '#fff'],
                ['offset' => 1, 'color' => $color]
            ]
        ];
    }, $data, array_keys($data));

    $seriesData = array_map(function($item, $index) use ($colorList) {
        return [
            'name' => $item['label'],
            'value' => $item['value'],
            'itemStyle' => [
                'color' => $colorList[$index]
            ]
        ];
    }, $data, array_keys($data));

    $type = $chart['chartType'] ?? 'gauge';

    $chartLegendPosition = $chart['chartLegendPosition'] ?? null;

    $orient = ($chartLegendPosition === 'left' || $chartLegendPosition === 'right') ? 'vertical' : 'horizontal';
    $topPosition = ($chartLegendPosition === 'left' || $chartLegendPosition === 'right') ? '10%' : $chartLegendPosition;
    $leftPosition = ($chartLegendPosition === 'left' || $chartLegendPosition === 'right') ? $chartLegendPosition : '0';

    $commonOptions = [
        'title' => [
            'text' => isset($chart['caption']) ? $chart['caption'] : ''
        ],
        'tooltip' => [
            'trigger' => 'item',
            'show' => !in_array($type, ['gauge', 'funnel'])
        ],
        'legend' => [
            'orient' => $orient,
            'left' =>  $leftPosition,
            'top' =>  $topPosition,
            'show' => ($chart['HideLegend'] ?? '0') !== '1',
            'type' => 'scroll',
            'pageButtonItemGap' => 10,
            'pageButtonGap' => 10,
            'pageIconColor' => '#999',
            'pageIconInactiveColor' => '#ccc',
            'pageIconSize' => 10,
            'formatter' => function($name) {
                return strlen($name) < 10 ? $name : substr($name, 0, 12) . '...';
            },
            'tooltip' => [
                'show' => true
            ]
        ]
    ];

    $specificOptions = [];

    switch ($type) {
        case 'pie':
            $specificOptions = [
                'series' => [
                    [
                        'type' => 'pie',
                        'radius' => '70%',
                        'center' => ['40%', '50%'],
                        'data' => $seriesData,
                        'animation' => false,
                        'label' => [
                            'show' => $chart['displayChartLabelFlag'] === '1'
                        ],
                        'itemStyle' => [
                            'shadowBlur' => 200,
                            'shadowColor' => 'rgba(0, 0, 0, 0.2)',
                            'normal' => [
                                'shadowBlur' => 20,
                                'shadowColor' => 'rgba(0, 0, 0, 0.5)',
                                'shadowOffsetX' => 10,
                                'shadowOffsetY' => 10
                            ]
                        ]
                    ]
                ]
            ];
            break;
        case 'funnel':
            $transformedData = array_map(function($item) {
                return [
                    'name' => $item['name'],
                    'value' => log(intval($item['value']) + 1)
                ];
            }, $seriesData);

            $maxValue = max(array_column($transformedData, 'value'));
            $normalizedData = array_map(function($item, $i) use ($seriesData, $maxValue) {
                return [
                    'name' => $seriesData[$i]['value'] . '-' . $item['name'],
                    'value' => ($item['value'] / $maxValue) * 100
                ];
            }, $transformedData, array_keys($transformedData));

            $specificOptions = [
                'series' => [
                    [
                        'type' => 'funnel',
                        'data' => $normalizedData,
                        'animation' => false,
                        'width' => '70%',
                        'label' => [
                            'show' => $chart['displayChartLabelFlag'] === '1',
                            'position' => 'inside'
                        ],
                        'gap' => 2,
                        'itemStyle' => [
                            'shadowBlur' => 10,
                            'shadowOffsetX' => 0,
                            'shadowOffsetY' => 0,
                            'shadowColor' => 'rgba(0, 0, 0, 0.5)'
                        ]
                    ]
                ]
            ];
            break;
        case 'line':

                // Prepare the array to store each series (line segment)
                $seriesArray = [];

                // Loop through the data and create individual series for each segment
                for ($i = 0; $i < count($data) - 1; $i++) {
                    $lineColor = !empty($chart['paletteColors']) 
                    ? '#' . $chart['paletteColors']
                    : ((array_key_exists('color', $data[$i]) && $data[$i]['color'] === '')
                        ? getRandomColor()
                        : '#' . $data[$i]['color']);

                    $seriesArray[] = [
                        'type' => 'line',
                        'data' => array_merge(array_fill(0, $i, null), [$data[$i]['value'], $data[$i + 1]['value']]),
                        'lineStyle' => [
                            'color' => $lineColor
                        ],
                        'itemStyle' => [
                            'color' => $lineColor
                        ],
                        'label' => [
                            'show' => $chart['displayChartLabelFlag'] === '1',
                            'position' => 'top',
                            'color' => '#000',
                            'formatter' => '{c}',
                            'fontSize' => 12
                        ]
                    ];
                }

                // Return the final chart options
                $specificOptions = [
                    'xAxis' => [
                        'type' => 'category',
                        'data' => array_column($data, 'label'),
                        'name' => isset($chart['XAxisName']) ? $chart['XAxisName'] : null
                    ],
                    'yAxis' => [
                        'type' => 'value',
                        'boundaryGap' => [0, '30%'],
                        'name' => isset($chart['YAxisName']) ? $chart['YAxisName'] : null
                    ],
                    'series' => $seriesArray
                ];
            break;
        case 'vbar':
            $specificOptions = [
                'xAxis' => [
                    'type' => 'category',
                    'data' => array_column($data, 'label'),
                    'name' => isset($chart['XAxisName']) ? $chart['XAxisName'] : ''
                ],
                'yAxis' => [
                    'type' => 'value',
                    'name' => isset($chart['YAxisName']) ? $chart['YAxisName'] : ''
                ],
                'grid' => [
                    'right' => 150
                ],
                'series' => [
                [
                    'type' => 'bar',
                    'animation' => false,
                    'label' => [
                        'show' => $chart['YAxisDisplayLabelMode'] === '1',
                        'position' => $chart['XAxisDisplayValuePosition'] === 'OUTSIDE' ? 'top' : 'insideTop'
                    ],
                    'data' => array_map(function($item, $index) use ($defaultColors) {
                        return [
                            'value' => $item['value'],
                            'itemStyle' => [
                                'color' => $defaultColors[$index] // Apply dynamic color for each bar
                            ]
                        ];
                    }, $data, array_keys($data)),
                ]
            ]

            ];
            break;
            case 'gauge':
                $gaugeColorRanges = array_map(function($color) use ($chart) {
                    return [
                        (float)$color['maxValue'] / (float)($chart['upperLimit'] ?? '100'),
                        [
                            'type' => 'radial',
                            'x' => 0.5,
                            'y' => 0.5,
                            'r' => 5,
                            'colorStops' => [
                                ['offset' => 0, 'color' => '#' . $color['code']],
                                ['offset' => 1, 'color' => '#000']
                            ]
                        ]
                    ];
                }, $rawData['colorRange']['color']);
                
                $specificOptions = [
                    'series' => [
                        [
                            'type' => 'gauge',
                            'radius' => isset($chart['gaugeOuterRadius']) && is_numeric($chart['gaugeOuterRadius'])
                                ? $chart['gaugeOuterRadius'] . '%'
                                : ($chart['gaugeOuterRadius'] ?? '70%'),
                            'startAngle' => $chart['gaugeStartAngle'] ?? 180,
                            'center' => ['50%', '60%'],
                            'endAngle' => $chart['gaugeEndAngle'] ?? 0,
                            'min' => 0,
                            'max' => (float)($chart['upperLimit'] ?? '100'),
                            'detail' => [
                                'show' => $chart['showValue'] == '1',
                                'formatter' => '{value}',
                                'color' => 'inherit',
                                'valueAnimation' => true,
                                'offsetCenter' => [0, '30%']
                            ],
                            'axisLabel' => [
                                'show' => $chart['valueBelowPivot'] == '1',
                                'color' => 'inherit',
                                'distance' => 40,
                                'rotate' => 'tangential',
                                'formatter' => function($value) {
                                    $max = 200;
                                    return ($value / $max * 10) % 2 === 0 ? $value : '';
                                }
                            ],
                            'axisLine' => [
                                'lineStyle' => [
                                    'width' => 30,
                                    'color' => $gaugeColorRanges
                                ]
                            ],
                            'splitLine' => [
                                'show' => true,
                                'distance' => -30,
                                'length' => 30,
                                'lineStyle' => [
                                    'color' => '#fff'
                                ]
                            ],
                            'axisTick' => [
                                'show' => true,
                                'distance' => -30,
                                'length' => 8,
                                'lineStyle' => [
                                    'color' => '#fff',
                                    'width' => 2
                                ]
                            ],
                            'pointer' => [
                                'length' => '70%',
                                'width' => 6,
                                'itemStyle' => [
                                    'color' => [
                                        'type' => 'radial',
                                        'x' => 0.5,
                                        'y' => 0.5,
                                        'r' => 15,
                                        'colorStops' => [
                                            ['offset' => 0, 'color' => '#C1232B'],
                                            ['offset' => 1, 'color' => '#000']
                                        ]
                                    ]
                                ]
                            ],
                            'title' => [
                                'show' => false
                            ],
                            'data' => !empty($rawData['dials']['dial']) ? array_map(function($item) {
                                return [
                                    'name' => $item['rearExtension'],
                                    'value' => $item['value']
                                ];
                            }, $rawData['dials']['dial']) : [['value' => 0, 'name' => 'No Data']]
                        ]
                    ]
                ];
                break;
        case 'donut-pie':
            $specificOptions = [
                'series' => [
                    [
                        'type' => 'pie',
                        'radius' => ['40%', '70%'],
                        'avoidLabelOverlap' => true,
                        'animation' => false,
                        'label' => [
                            'show' => isset($chart['displayChartLabelFlag']) && $chart['displayChartLabelFlag'] == '1',
                            'position' => 'center'
                        ],
                        'emphasis' => [
                            'label' => [
                                'show' => isset($chart['displayChartLabelFlag']) && $chart['displayChartLabelFlag'] == '1',
                                'fontSize' => '15'
                            ]
                        ],
                        'labelLine' => [
                            'show' => false
                        ],
                        'data' => count($seriesData) > 0 ? $seriesData : [['value' => 0, 'name' => 'No Data', 'itemStyle' => ['color' => '#91cc75']]]
                    ]
                ]
            ];
            break;
        default:
            throw new Exception("Unsupported chart type: " . (isset($chart['chartType']) ? $chart['chartType'] : ''));
    }

    return array_merge($commonOptions, $specificOptions);
}


function getRandomColor() {
    $letters = '0123456789ABCDEF';
    $color = '#';
    for ($i = 0; $i < 6; $i++) {
        $color .= $letters[rand(0, 15)];
    }
    return $color;
}

?>
